#!/usr/bin/env python

__progname__ = "Deftool"
__version__ = "VERSION"
__author__ = "Anders Bruun Olsen <anders@bruun-olsen.net>"
__copyright__ = """Copyright (c) 2004 Anders Bruun Olsen.

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA."""


__usage__ = """%prog [-a archiveurl] [-d deffile] [-c classfile] [-D [-r]] <comic>"""

import re
import os
import readline
import urllib2
import sys
import traceback
import ConfigParser
import time
from optparse import OptionParser, OptionValueError
from urlparse import urljoin
from string import replace


def check_verbose(option, opt, value, parser, **kwargs):
	if getattr(parser.values, option.dest):
		raise OptionValueError("Can't be both verbose and quiet.")
	else:
		setattr(parser.values, option.dest, kwargs["val"])


def check_action(option, opt, value, parser, **kwargs):
	if getattr(parser.values, option.dest):
		raise OptionValueError("Can't do more than one action at a time.")
	else:
		setattr(parser.values, option.dest, kwargs["val"])


def green(text):
	"""Return green text."""
	return "\x1b[32;01m" + text + "\x1b[0m"


def red(text):
	"""Return red text."""
	return "\x1b[31;01m" + text + "\x1b[0m"


def blue(text):
	"""Return red text."""
	return "\x1b[34;01m" + text + "\x1b[0m"
	
	
def main():
	parser = OptionParser( usage=__usage__, version=__progname__ + " " + __version__ )
	parser.add_option("-a", "--archiveurl", action="store", type="string", dest="archiveurl", help="Set custom archiveurl.")
	parser.add_option("-d", "--deffile", action="store", type="string", dest="deffile", help="Set custom deffile to use.")
	parser.add_option("-c", "--classfile", action="store", type="string", dest="classfile", help="Set custom classfile to use.")
	parser.add_option("-u", "--useragent", action="store", type="string", dest="useragent", help="Set custom useragent.")
	parser.add_option("-D", "--download", action="store_true", default=False, dest="download", help="Download files. If this is not set, the files won't be downloaded.")
	parser.add_option("-r", "--redownload", action="store_true", default=False, dest="redownload", help="Redownload files if they already exist locally.")
	(options, args) = parser.parse_args()

	try:
		comic = Comic(options, args[0])
	except KeyboardInterrupt:
		sys.stderr.write("\nProgram interrupted by user.\n")
		sys.exit(10)
	except urllib2.HTTPError, error:
		sys.stderr.write( "\nA HTTP error occurred: " + str(error.code) + ": " + str(error.msg) + ", when trying to access " + str(error.geturl()) + "\n" )
		sys.exit(20)
	except DownloadErrors, error:
		sys.stderr.write("\nThe following errors occurred:\n")
		for err in error.errors:
			sys.stderr.write("[" + err[0] + "] When downloading " + err[1] + ": " + err[2] + "\n")
		sys.exit(30)
	except:
		sys.stderr.write("\nAn unhandled exception occurred. Please report this problem to http://bugs.skumleren.net along with this traceback:\n")
		traceback.print_exc(file=sys.stderr)
		sys.exit(42)




class deftool:
	def __init__(self):
		self.name = raw_input("Name of webcomic: ")
		self.authors = raw_input("Name of author(s): ")
		self.homepage = raw_input("Homepage: ")
		self.comictype = raw_input("Type (search or lateststatic: ")
		while self.comictype != "search" and self.comictype != "lateststatic":
			self.comictype = raw_input("Type (search or lateststatic: ")
		if self.comictype == "search":
			self.type_search()
		elif self.comictype == "lateststatic":
			self.type_lateststatic()

	def type_search(self):
		self.frontpageurl = raw_input("Frontpageurl: ")
		self.frontpageexp = raw_input("Frontpageexp: ")
		if self.frontpageexp != "":
			self.test_frontpageexp()
			html = urllib2.urlopen(self.homepage).read()
			self.fullfrontpageurl = self.frontpageurl + re.findall(self.frontpageexp, html)[0]
		else:
			self.fullfrontpageurl = self.frontpageurl
		self.imgurl = raw_input("Imgurl: ")
		self.imgexp = raw_input("Imgexp: ")
		self.test_imgexp()
		self.previousurl = raw_input("Previousurl: ")
		self.previousexp = raw_input("Previousexp: ")
		self.test_previousexp()
		self.test_archive()
		self.print_search()

	def type_lateststatic(self):
		print "#FIXME"

	def test_frontpage(self):
		html = urllib2.urlopen(self.homepage).read()
		matches = re.findall(self.frontpageexp, html)
		print "Matches: " + str(matches)
		newfrontpageexp = raw_input("New frontpageexp: ")
		if newfrontpageexp != "":
			self.frontpageexp = newfrontpageexp
			self.test_frontpage()

	def test_imgexp(self):
		html = urllib2.urlopen(self.fullfrontpageurl).read()
		matches = re.findall(self.imgexp, html)
		print "Matches: " + str(matches)
		newimgexp = raw_input("New imgexp: ")
		if newimgexp != "":
			self.imgexp = newimgexp
			self.test_imgexp()
	
	def test_previousexp(self):
		html = urllib2.urlopen(self.fullfrontpageurl).read()
		matches = re.findall(self.previousexp, html)
		print "Matches: " + str(matches)
		newpreviousexp = raw_input("New previousexp: ")
		if newpreviousexp != "":
			self.previousexp = newpreviousexp
			self.test_previousexp()

	def test_archive(self):
		print "Going back through 5 strips in archive."
		html = urllib2.urlopen(self.fullfrontpageurl).read()
		print self.fullfrontpageurl + ": ",
		for x in range(5):
			img = re.findall(self.imgexp, html)
			print "Imgs: " + str(img)
			url = re.findall(self.previousexp, html)[0]
			print "Prev: " + url
			html = urllib2.urlopen(url).read()

	def print_search(self):
		print "[" + self.name + "]"
		print "type = " + self.type
		print "authors = " + self.authors
		print "homepage = " + self.homepage
		print "frontpageurl = " + self.frontpageurl
		if self.frontpageexp != "":
			print "frontpageexp = " + self.frontpageexp
		print "imgurl = " + self.imgurl
		print "imgexp = " + self.imgexp
		print "previousurl = " + self.previousurl
		print "previousexp = " + self.previousexp


class Comic:
	"""Comic class."""
	def __init__(self, options, name):
		self.options = options
		if self.options.deffile:
			self.deffile = self.options.deffile
		else:
			self.deffile = "/etc/collector/comics.def"
		if self.options.classfile:
			self.classfile = self.options.classfile
		else:
			self.classfile = "/etc/collector/comics.classes"
		self.config = ConfigParser.ConfigParser()
		self.config.read(os.path.expanduser("~/.collector/collector.cfg"))
		self.dir = self.config.get("Main", "comicsdir")
		if self.dir[-1] != "/":
			self.dir += "/"
		self.dir += name + "/"

		if self.options.useragent:
			self.useragent = useragent
		else:
			self.useragent = "WebcomicsCollector/Deftool"

		self.comicsdef = ConfigParser.ConfigParser()
		self.comicsdef.read(self.deffile)
		self.comicsclasses = ConfigParser.ConfigParser()
		self.comicsclasses.read(self.classfile)
		
		options = {}
		if self.comicsdef.has_option(name, "class"):
			try:
				for opt in self.comicsclasses.items( self.comicsdef.get(name, "class") ):
					options[opt[0]] = opt[1]
			except ConfigParser.NoSectionError, error:
				raise ConfigurationError("The definition references non-existing class '" + error.section + "'")
		for opt in self.comicsdef.items(name):
			options[opt[0]] = opt[1]
		for opt in options:
			matches = None
			while matches != []:
				matches = re.findall( "\{\{(.+?)\}\}", options[opt] )
				for match in matches:
					if match == "ext":
						options[opt] = re.sub( "\{\{" + match + "\}\}", "(?:jpg|jpeg|jpe|gif|png|bmp)", options[opt])
					elif match == "signs":
						options[opt] = re.sub( "\{\{" + match + "\}\}", "[a-z0-9_\-\.%]+?", options[opt])
					elif match == "filename":
						options[opt] = re.sub( "\{\{" + match + "\}\}", "[a-z0-9_\-\.%]+?\.(?:jpg|jpeg|jpe|gif|png|bmp)", options[opt])
					elif opt == "archiveurl" and match[:5] == "DATE:":
						date = time.strftime(match[5:])
						url = re.sub("\{\{" + match + "\}\}", date, options[opt])
						try:
							urllib2.urlopen(url)
						except urllib2.HTTPError:
							date = time.strftime(match[5:], time.localtime(time.time() - 86400))
							url = re.sub("\{\{" + match + "\}\}", date, options[opt])
						options[opt] = url
					else:
						try:
							options[opt] = re.sub( "\{\{" + match + "\}\}", options[match], options[opt])
						except KeyError:
							raise ConfigurationError("Invalid substitution '{{" + match + "}}' in " + opt + ".")
		for opt in options:
			setattr(self, opt, options[opt])
			print str(opt) + ": " + str(options[opt])

		if self.options.download and not os.path.isdir(self.dir):
			self.log("'" + self.dir + "' does not exist, now creating.")
			os.mkdir(self.dir)

		self.checkDef()
		self.log(green(" * ") + name, tab = False)
		if self.type == "search":
			self.fetchArchive()
		elif self.type == "lateststatic":
			self.fetchLatestStatic()
		else:
			raise ConfigurationError("Invalid type '" + self.type + "' specified.")
			
	def checkDef(self):
		# Check type option
		try:
			self.type
		except AttributeError:
			raise MissingOption("type")
		if self.type != "search" and self.type != "lateststatic":
			raise ConfigurationError("Invalid type '" + self.type + "' specified for comic '" + self.name + "'")

		# Check homepage option
		try:
			self.homepage
		except AttributeError:
			raise MissingOption("homepage")

		# Check archiveurl option
		try:
			self.archiveurl = str(self.archiveurl)
		except:
			self.archiveurl = self.homepage
		if self.options.archiveurl:
			self.log("Archiveurl is " + green(self.options.archiveurl))
			self.archiveurl = self.options.archiveurl
			try:
				del self.archiveexp
			except AttributeError:
				pass

		# Check authors option
		try:
			self.authors = str(self.authors)
		except AttributeError:
			raise MissingOption("authors")

		# Check imgexp option
		try:
			if self.type == "search":
				self.imgexp = re.compile(self.imgexp, re.I)
		except re.error, err:
			raise RegExpError("imgexp", self.imgexp, err)
		except AttributeError:
			raise MissingOption("imgexp")

		# Check imgurl option
		try:
			if self.type == "lateststatic":
				self.imgurl
		except AttributeError:
			raise MissingOption("imgurl")
		
		# Check altimgexp option
		try:
			if self.type == "search" and self.altimgexp:
				self.altimgexp = re.compile(self.altimgexp, re.I)
		except re.error, err:
			raise RegExpError("altimgexp", self.altimgexp, err)
		except AttributeError:
			pass

		# Check previousexp option
		try:
			if self.type == "search":
				self.previousexp = re.compile(self.previousexp, re.I)
		except re.error, err:
			raise RegExpError("previousexp", self.previousexp, err)
		except AttributeError:
			raise MissingOption("previousexp")
			
		# Check altpreviousexp option
		try:
			if self.type == "search" and self.altpreviousexp:
				self.altpreviousexp = re.compile(self.altpreviousexp, re.I)
		except re.error, err:
			raise RegExpError("altpreviousexp", self.altpreviousexp, err)
		except AttributeError:
			pass
	
	def log(self, text = "", newline = True, tab = True):
		"""Log text to a logfile (#FIXME) depending on loglevel and depending upon verboselevel print to screen also."""
		if tab:
			text = "\t" + text
		if newline:
			print text
		else:
			print text,

	def downloadImage(self, files, referer):
		"""
		Download image(s) and places it/them in comicdir.
		Takes a list of two-part tuples containing url and name.
		Does not return anything but raises exceptions of anything goes wrong.
		"""
		for f in files:
			url, name = f
			if not os.path.isfile(self.dir + name) or self.options.redownload:
				# Download file if it does not exist or if redownload option is set
				request = urllib2.Request(url)
				request.add_header("Referer", referer)
				request.add_header("User-Agent", self.useragent)
				try:
					self.log("Downloading file: " + name)
					img = urllib2.urlopen(request).read()
				except urllib2.HTTPError, error:
					raise DownloadError(str(error.geturl()), "[" + str(error.code) + "] " + str(error.msg))
				open(self.dir + name, "w").write(img)
			else:
				self.log("Not downloading file (already downloaded): " + name)

	def applyArchiveExp(self):
		"""If archiveexp exists, archiveurl will be searched with it and the result joined with archiveurl."""
		try:
			self.archiveexp
			self.log("Finding archiveurl:", newline = False)
			request = urllib2.Request(self.archiveurl)
			request.add_header("User-Agent", self.useragent)
			try:
				html = urllib2.urlopen(request).read()
			except urllib2.HTTPError, error:
				raise ArchiveURLError( "[HTTP Error " + str(error.code) + "] " + str(error.msg) + ", when trying to find archiveurl on " + str(error.geturl()) )
			html = re.sub("\n|\r|\t", "", html)
			html = re.sub(">\s+<", "><", html)
			matches = re.search( self.archiveexp, html )
			if matches is not None:
				self.archiveurl = urljoin(self.archiveurl, matches.groups()[0])
				self.log("'" + self.archiveurl + "'", tab = False)
				del self.archiveexp
			else:
				raise ArchiveURLError( "Could not find '" + self.archiveexp + "' on " + self.archiveurl )
		except AttributeError:
			pass

	def applyImgExp(self, html, url):
		"""Search specified html using the imgexp regexp and return the results."""
		strips = re.findall(self.imgexp, html)
		if strips == []:
			# No strips found, try altimgexp instead
			try:
				strips = re.findall(self.altimgexp, html)
				self.log(blue("(using altimgexp) "), tab=False, newline=False)
			except AttributeError:
				pass
		# Create full URLs for all stripfiles
		for n in range(len(strips)):
			try:
				strips[n] = self.imgurl + str(strips[n])
			except AttributeError:
				strips[n] = urljoin(url, strips[n])
		# Find stripnames
		try:
			names = re.findall(self.imgname, html)
		except AttributeError:
			names = []
			for strip in strips:
				try:
					self.pathlevels
					head, tail = os.path.split(strip)
					name = tail
					for n in range(int(self.pathlevels) - 1):
						head, tail = os.path.split(head)
						name = tail + "_" + name
				except AttributeError:
					name = os.path.basename(strip)
				names.append(name)
		returnlist = []
		for n in range(len(strips)):
			name = replace(names[n], "/", "_")
			returnlist.append( (strips[n], name) )
		return returnlist

	def applyPreviousExp(self, html, url):
		"""Search specified html using the previousexp regexp and return the results."""
		matches = re.findall(self.previousexp, html)
		if matches == []:
			# No matches found, try altpreviousexp instead
			try:
				matches = re.findall(self.altpreviousexp, html)
				self.log(blue("using altpreviousexp on " + url))
			except AttributeError:
				pass
		if matches != []:
			prevstrip = urljoin(url, str(matches[0]))
		else:
			prevstrip = ""
		return prevstrip

	def fetchLatestStatic(self):
		# Find archiveurl if needed
		self.applyArchiveExp()
		self.log("Fetching lateststatic.")
		# Find name of latest downloaded strip
		if os.path.isfile(self.dir + "strips"):
			lateststripname = open(self.dir + "strips").readlines().pop().strip()
		else:
			lateststripname = None

		if lateststripname is not None:
			lateststrip = open(self.dir + lateststripname).read()
		else:
			lateststripname = 0
			lateststrip = ""
		newstripname = lateststripname + 1
		request = urllib2.Request(self.imgurl)
		request.add_header("Referer", self.archiveurl)
		request.add_header("User-Agent", self.useragent)
		try:
			img = urllib2.urlopen(request).read()
		except urllib2.HTTPError, error:
		
			raise DownloadError( str(error.geturl()), "[" + str(error.code) + "] " + str(error.msg) )
		if img != lateststrip and self.options.download:
			open(self.dir + str(newstripname), "w").write(img)
			open(self.dir + "strips", "a").write(newstripname + "\n")
		elif self.options.download and self.options.redownload:
			open(self.dir + str(lateststripname), "w").write(img)

	def fetchArchive(self):
		"""Download the archived strips."""
		self.applyArchiveExp()
		self.log("Fetching archive:")
		self.log("Finding strip(s):", newline = False)
		# Grab page
		request = urllib2.Request(self.archiveurl)
		request.add_header("User-Agent", self.useragent)
		try:
			html = urllib2.urlopen(request).read()
		except urllib2.HTTPError, error:
			raise DownloadError( str(error.geturl()), "[" + str(error.code) + "] " + str(error.msg) )
		# Remove tabs and newlines
		html = re.sub("\n|\r|\t", "", html)
		html = re.sub(">\s+<", "><", html)
		# Find neweststrip(s)
		neweststrips = self.applyImgExp(html, self.archiveurl)
		if neweststrips == []:
			self.log()
			self.log(red("No strips found on " + self.archiveurl))
		else:
			# Log the names found.
			self.log(neweststrips[0][1], newline = False, tab = False)
			for strip in neweststrips[1:]:
				self.log(", " + strip[1], newline = False, tab = False)
			self.log()
		visitedurls = [self.archiveurl]
		# Find url to previous strip
		currenturl = self.applyPreviousExp(html, self.archiveurl)
		if currenturl == "":
			print "Can't find previousexp on " + self.archiveurl
			sys.exit(1)
		# Make concatname for use in strips file
		concatname = neweststrips[0][1]
		for strip in neweststrips[1:]:
			concatname += "|" + strip[1]
		stripnames = [concatname]
		# Make stripurl.db list
		stripurls = []
		for strip in neweststrips:
			stripurls.append( (strip[1], self.archiveurl) )

		if self.options.download:
			self.downloadImage(neweststrips, referer = self.archiveurl)
		prevurl = self.archiveurl
		try:
			while 1:
				self.log("Finding strip(s) on " + currenturl + " :", newline = False)
				request = urllib2.Request(currenturl)
				request.add_header("Referer", prevurl)
				request.add_header("User-Agent", self.useragent)
				try:
					html = urllib2.urlopen(request).read()
				except urllib2.HTTPError, error:
					raise DownloadError( str(error.geturl()), "[" + str(error.code) + "] " + str(error.msg) )
				html = re.sub("\n|\r|\t", "", html)
				html = re.sub(">\s+<", "><", html)
				strips = self.applyImgExp(html, currenturl)
				if strips == []:
					self.log()
					self.log(red("No strips found on " + currenturl))
				else:
					concatname = strips[0][1]
					self.log(strips[0][1], newline = False, tab = False)
					for strip in strips[1:]:
						concatname += "|" + strip[1]
						self.log(", " + strip[1], newline = False, tab = False)
					self.log()
					stripnames.append(concatname)
				if self.options.download:
					self.downloadImage(strips, referer = currenturl)
				for strip in strips:
					stripurls.append( (strip[1], currenturl) )
				newurl = self.applyPreviousExp(html, currenturl)
				visitedurls.append(currenturl)
				if newurl == "":
					break
				elif newurl in visitedurls:
					self.log(red("Loop detected, stopping download"))
					break
				else:
					prevurl = currenturl
					currenturl = newurl
		except KeyboardInterrupt:
			print "Keyboard Interrupt detected, stopping archive download."
		except ImgExpError, error:
			print error.errormsg + " Stopping archive download."
		if self.options.download:
			self.log("Writing '" + self.dir + "strips'")
			stripnames.reverse()
			striplist = ""
			for name in stripnames:
				striplist += name + "\n"
			open(self.dir + "strips", "w").write(striplist)


class RegExpError(Exception):
	"""Exception raised if a regular expression contains any errors."""
	def __init__(self, option, exp, experror):
		self.option = option
		self.exp = exp
		self.experror = str(experror)
	def __str__(self):
		return "The regexp in " + self.option + ", " + self.exp + ", contains this error: " + self.experror


class DownloadError(Exception):
	"""Exception raised if a file could not be downloaded."""
	def __init__(self, url, errormsg):
		self.url = url
		self.errormsg = errormsg
	def __str__(self):
		return self.url + " : " + self.errormsg


class ConfigurationError(Exception):
	"""Exception raised if a configuration error has been encountered."""
	def __init__(self, errormsg):
		self.errormsg = errormsg
	def __str__(self):
		return self.errormsg


class MissingOption(Exception):
	"""Exception raised if a required option in a def is missing."""
	def __init__(self, option):
		self.option = option
	def __str__(self):
		return "The required option " + self.option + " is missing from comic definition."


class InvalidUrlError(Exception):
	"""Exception raised if a URL specified in a def is invalid."""
	def __init__(self, option, url):
		self.option = option
		self.url = url
	def __str__(self):
		return "Option " + self.option + " contains invalid URL: " + self.url


class DownloadErrors(Exception):
	"""Exception raised if there was any download errors."""
	def __init__(self, errors):
		self.errors = errors
	def __str__(self):
		errormsg = ""
		for error in self.errors:
			errormsg += error + "\n"
		return errormsg


class ComicErrors(Exception):
	"""Exception raised at the end of a run if any errors occurred."""
	def __init__(self, errors):
		self.errors = errors
	def __str__(self):
		errormsg = ""
		for error in self.errors:
			errormsg += error[0] + ":\n"
			errormsg += str(error[1]) + "\n\n"
		return errormsg


if __name__ == "__main__":
	main()
