#!/usr/bin/env python

import md5
import cPickle
import ConfigParser
import os.path
import gdbm
import re
from optparse import OptionParser, OptionValueError


def main():
	parser = OptionParser(usage="%prog <comicname>")
	(options, args) = parser.parse_args()
	comicsdir, lateststaticcomics = findLatestStaticComics()
	comicslist = []
	if args != []:
		for comic in args:
			if comic in lateststaticcomic:
				comicslist.append(comic)
			else:
				print comic + " is not a lateststatic comic in your comicslist, not fixing."
	else:
		comicslist = lateststaticcomics
	for comic in comicslist:
		dir = comicsdir + comic + "/"
		renameStrips(dir, readStripsList(dir))
		md5sumStrips(dir, readStripsList(dir))


def findLatestStaticComics():
	print "Finding lateststatic comics"
	cfgdir = os.path.expanduser("~/.collector/")
	config = ConfigParser.ConfigParser()
	config.read(cfgdir + "collector.cfg")
	comicsclasses = ConfigParser.ConfigParser()
	if os.path.isfile("/etc/collector/comics.classes"):
		comicsclasses.read("/etc/collector/comics.classes")
	if os.path.isfile("/usr/local/etc/collector/comics.classes"):
		comicsclasses.read("/usr/local/etc/collector/comics.classes")
	if os.path.isfile(cfgdir + "comics.classes"):
		comicsclasses.read(cfgdir + "comics.classes")
	# Load comics definitions
	comicsdef = ConfigParser.ConfigParser()
	if os.path.isfile("/etc/collector/comics.def"):
		comicsdef.read("/etc/collector/comics.def")
	if os.path.isfile("/usr/local/etc/collector/comics.def"):
		comicsdef.read("/usr/local/etc/collector/comics.def")
	if os.path.isfile(cfgdir + "comics.def"):
		comicsdef.read(cfgdir + "comics.def")
	comicsdir = os.path.expanduser(config.get("Main", "comicsdir"))
	if comicsdir[-1] != "/":
		comicsdir += "/"
	comicslist = []
	for comic in config.get("Main", "comics").split(","):
		comic = comic.strip()
		if comic != "" and comicsdef.has_section(comic):
			if comicsdef.has_option(comic, "type") and comicsdef.get(comic, "type") == "lateststatic":
				comicslist.append(comic)
			elif comicsdef.has_option(comic, "class") and comicsclasses.get(comicsdef.get(comic, "class"), "type") == "lateststatic":
				comicslist.append(comic)
	return comicsdir, comicslist


def readStripsList(dir):
	print "Reading " + dir + "strips"
	returnList = []
	strips = open(dir + "strips", "r")
	for strip in strips:
		returnList.append(strip.strip())
	return returnList


def md5sumStrips(dir, strips):
	print "Making md5sums for strips in " + dir
	md5sums = []
	for strip in strips:
		for file in strip.split("|"):
			print "Creating md5sum for " + file
			sum = md5.new()
			sum.update(open(dir + file, "r").read())
			md5sums.append( (file, sum.hexdigest()) )
	print "Writing " + dir + "strips.md5"
	cPickle.dump(md5sums, open(dir + "strips.md5", "w"), 2)


def renameStrips(dir, strips):
	print "Renaming strips"
	stripurls = gdbm.open(dir + "stripurls.db", "c")
	stripslist = ""
	exp = re.compile("\.(?:gif|jpg|jpe|jpeg|png)", re.I)
	for file in strips:
		newfile = exp.sub("", file)
		if newfile != file:
			print file + " -> " + newfile
			os.rename(dir + file, dir + newfile)
			stripslist += newfile + "\n"
			try:
				stripurls[newfile] = stripurls[file]
				del stripurls[file]
			except KeyError:
				pass
	stripurls.close()
	print "Writing " + dir + "strips"
	open(dir + "strips", "w").write(stripslist)


if __name__ == "__main__":
	main()
